#!/usr/bin/env python3

# Author: Ben Mezger <me@benmezger.nl>
# Created at <2023-09-23 Sat 21:35>


class RetryExaustedError(BaseException):
    """Raised when the number of retries was exausted"""

    ...


class RetryStrategyExausted(BaseException):
    """Raised when the strategy was exausted"""

    ...



#!/usr/bin/env python3

# Author: Ben Mezger <me@benmezger.nl>
# Created at <2023-09-23 Sat 21:35>

import abc
import functools
import inspect
import logging
import typing as t
from dataclasses import dataclass
from datetime import datetime

from pyretries.exceptions import RetryExaustedError, RetryStrategyExausted
from pyretries.strategy import Strategy

ConditionT = t.TypeVar("ConditionT")
ReturnT = t.TypeVar("ReturnT")
FuncT = t.Callable[..., ReturnT]
AfterHookFuncT = t.Callable[[Exception | ReturnT], None]
BeforeHookFuncT = t.Callable[..., None]
RetryExceptionCallHook = t.Callable[[Exception], None]


_logger = logging.getLogger(__name__)


@dataclass
class RetryState(t.Generic[ReturnT]):
    """Stores current retry state

    Args:
        func: Function address to retry
        args: `func` non-positional arguments
        kwargs: `func` positional arguments
        start_time: Timestamp when retry first started
        end_time: Timestamp when retry ended
        strategy_func: Next strategy to run
        current_attempts: Number of current retry attempts
        exception: Exception raised by `func`

    """

    func: FuncT[ReturnT]
    start_time: int
    end_time: int = 0
    strategy: Strategy[ReturnT] | None = None
    args: t.Sequence[t.Any] | None = None
    kwargs: t.Dict[str, t.Any] | None = None
    current_attempts: int = 0
    returned_value: ReturnT | None = None
    exception: Exception | None = None

    @property
    def raised(self) -> bool:
        """Checks if `func` raised an exception"""
        return isinstance(self.exception, Exception)

    def clear(self) -> None:
        """Clears retry state"""
        self.exception = None
        self.returned_value = None

    def __repr__(self) -> str:
        cls_name = type(self).__name__
        return (
            f"{cls_name}(start_time={self.start_time}, "
            f"end_time={self.end_time}, "
            f"current_attempts={self.current_attempts}, "
            f"exception={repr(self.exception)}), "
            f"returned_value={self.returned_value})"
        )


class BaseRetry(abc.ABC, t.Generic[ReturnT]):
    """
    Base class for all retry implementations.
    Requires `__call__` implementation

    Examples:
        >>> Class RetryExample(BaseRetry[ReturnT]):
                ...
                def __call__(
                    self,
                    func: FuncT[ReturnT],
                    *args: t.Tuple[t.Any],
                    **kwargs: t.Dict[t.Any, t.Any],
                ) -> ReturnT | Exception | None:
                     ...

    """

    on_exceptions: set[type[Exception]] | None

    def __init__(
        self,
        strategies: t.Sequence[Strategy[ReturnT]] = [],
        on_exceptions: t.Sequence[type[Exception]] | None = None,
        before_hooks: t.Sequence[BeforeHookFuncT] | None = None,
        after_hooks: t.Sequence[AfterHookFuncT[ReturnT]] | None = None,
        retry_exception_hook: RetryExceptionCallHook | None = None,
        should_log: bool = True,
    ) -> None:
        """
        Args:
            strategies: Sequence of retry strategies
            on_exceptions: Sequence of exceptions to apply a retry strategy.
            before_hooks: Hooks to run before running `func`. Runs Before strategy.
            after_hooks: Hooks to run after running `func`. Runs before strategy.
            retry_exception_hook: Hook to run when `func` raised an exception. Runs before strategy.
            should_log: Specifies whether retry should log actions
        """
        self.strategies = list(reversed(strategies))
        self.on_exceptions = set(on_exceptions or []) or None

        self.before_hooks = before_hooks or []
        self.after_hooks = after_hooks or []
        self.retry_exception_hook = retry_exception_hook
        self.should_log = should_log

    @abc.abstractmethod
    def __call__(
        self,
        func: FuncT[ReturnT],
        *args: t.Tuple[t.Any],
        **kwargs: t.Dict[t.Any, t.Any],
    ) -> ReturnT | Exception | None:
        """
        Executes `func` and applies strategies

        Args:
            func: Address to function
            args: `func` non-positional arguments
            kwargs: `func` positional arguments

        """
        raise NotImplementedError

    def save_state(self, state: RetryState[ReturnT]) -> None:
        """
        Saves retry state to `func`

        Args:
            state: Current retry state
        """
        setattr(state.func, "state", state)

    def exec_strategy(self, state: RetryState[ReturnT]):
        """
        Applies user defined strategies.

        Args:
            state: Current retry state

        Raises:
            RetryExaustedError: Raised when strategy is exausted of no strategy is available
        """
        if state.strategy is None:
            if len(self.strategies):
                state.strategy = self.strategies.pop()
            else:
                raise RetryExaustedError

        try:
            if state.strategy.should_stop:
                raise RetryStrategyExausted

            if self.should_log:
                _logger.info(
                    f"Executing '{state.strategy.__class__.__name__}' retry strategy. "
                    f"Current attempt {state.current_attempts}"
                )

            state.strategy.eval(state.returned_value)
            state.current_attempts += 1

            if state.strategy.should_stop:
                state.strategy = None

        except RetryStrategyExausted:
            raise RetryExaustedError

    def apply(self, state: RetryState[ReturnT]) -> bool:
        """
        Checks if last state raised an exception and executes the next available strategy

        Raise `RetryExaustedError` if last executes strategy raised

        Args:
            state: Current retry state

        Raises:
            RetryExaustedError: Raised when `func` exception is not defined in `on_exceptions` sequence
                                or when strategy raises
        """
        try:
            if not state.raised:
                return False

            if (exc := state.exception.__class__) not in (self.on_exceptions or [exc]):
                raise RetryExaustedError from state.exception

            self.exec_strategy(state)
            state.clear()
            return True

        except RetryExaustedError as err:
            raise err from state.exception if state.raised else None

    def _pre_exec(self, _: RetryState[ReturnT]) -> None:
        """
        Should be called before running `func`

        Args:
            state: Current retry state. Currently not used.
        """
        for hook in self.before_hooks:
            hook()

    def _post_exec(
        self, state: RetryState[ReturnT], exception: Exception | None
    ) -> None:
        """
        Should be called after running `func`.
        If exception was raised, this should be passed in `exception` argument.

        Args:
            state: Current retry state. Currently not used.
            exception: Raised exception
        """
        if exception:
            state.exception = exception

            if self.retry_exception_hook:
                self.retry_exception_hook(exception)

        for hook in self.after_hooks:
            hook(state.exception or state.returned_value)

        state.end_time = int(datetime.now().timestamp())


class AsyncRetry(BaseRetry[ReturnT]):
    """
    Asynchronous retry

    Examples:
        >>> async def ok() -> bool:
        ...    return True
        >>> retry = AsyncRetry[t.Awaitable[bool]](strategies=[StopAfterAttemptStrategy(20)])
        >>> print(await retry(ok))
    """

    async def exec(self, state: RetryState[ReturnT]) -> None:
        """
        Executes `func` from `state`

        Args:
            state: Current retry state

        """
        assert inspect.iscoroutinefunction(
            state.func
        ), f"{self.__class__.__name__} needs an awaitable func"

        self._pre_exec(state)

        exception: Exception | None = None
        try:
            state.returned_value = await state.func(
                *(state.args or ()), **(state.kwargs or {})
            )
        except Exception as err:
            exception = err

        self._post_exec(state, exception)

    # NOTE: currently, the abtract method is a sync, and here we are defining async.
    # Pyright will alert about this, so we ignore the type for now until we
    # have a fix for it
    async def __call__(  # type:ignore
        self, func: FuncT[ReturnT], *args: t.Any, **kwargs: t.Any
    ) -> ReturnT | Exception | None:
        """
        Executes `func` and applies strategies

        Args:
            func: Address to function
            args: `func` non-positional arguments
            kwargs: `func` positional arguments

        Returns:
           Either `func`'s return value or None
        """
        state = RetryState[ReturnT](
            func=func,
            start_time=int(datetime.now().timestamp()),
            args=args,
            kwargs=kwargs,
        )

        if self.should_log:
            _logger.info(f"Calling '{func.__name__}'")

        should_reapply = True
        while should_reapply:
            await self.exec(state)

            if not (should_reapply := self.apply(state)):
                break

        self.save_state(state)
        return state.returned_value


class Retry(BaseRetry[ReturnT]):
    """
    Synchronous retry

    Examples:
        >>> def ok() -> bool:
        ...    return True
        >>> retry = Retry[bool](strategies=[StopAfterAttemptStrategy(20)])
        >>> print(retry(ok))
    """

    def exec(self, state: RetryState[ReturnT]) -> None:
        """
        Executes `func` from `state`

        Args:
            state: Current retry state

        """
        self._pre_exec(state)

        exception: Exception | None = None
        try:
            state.returned_value = state.func(
                *(state.args or ()), **(state.kwargs or {})
            )
        except Exception as err:
            exception = err

        self._post_exec(state, exception)

    def __call__(
        self, func: FuncT[ReturnT], *args: t.Any, **kwargs: t.Any
    ) -> ReturnT | Exception | None:
        """
        Executes `func` and applies strategies

        Args:
            func: Address to function
            args: `func` non-positional arguments
            kwargs: `func` positional arguments

        Returns:
           Either `func`'s return value or None
        """
        state = RetryState[ReturnT](
            func=func,
            start_time=int(datetime.now().timestamp()),
            args=args,
            kwargs=kwargs,
        )

        if self.should_log:
            _logger.info(f"Calling '{func.__name__}'")

        should_reapply = True
        while should_reapply:
            self.exec(state)

            if not (should_reapply := self.apply(state)):
                break

        self.save_state(state)
        return state.returned_value


def retry(
    strategies: t.Sequence[Strategy] = [],
    on_exceptions: t.Sequence[type[Exception]] | None = None,
    before_hooks: t.Sequence[BeforeHookFuncT] | None = None,
    after_hooks: t.Sequence[AfterHookFuncT[ReturnT]] | None = None,
    retry_exception_hook: RetryExceptionCallHook | None = None,
    should_log: bool = True,
):
    """
    Retry decorator. Works both for sync and async functions

    Examples:
        >>> @retry(strategies=[strategy.NoopStrategy(1)])
        ... def ok() -> bool:
        ...     return True
        >>> ok()
        INFO:retries.retry:Calling 'ok'
        True

    Args:
        strategies: Sequence of retry strategies
        on_exceptions: Sequence of exceptions to apply a retry strategy.
        before_hooks: Hooks to run before running `func`. Runs Before strategy.
        after_hooks: Hooks to run after running `func`. Runs before strategy.
        retry_exception_hook: Hook to run when `func` raised an exception. Runs before strategy.
        should_log: Specifies whether retry should log actions

    Returns:
        func (FuncT[ReturnT]): Functions return value or exception
    """

    def decorator_retry(
        func: FuncT[ReturnT],
    ) -> FuncT[ReturnT]:
        @functools.wraps(func)
        def wrapper_retry(*args, **kwargs):
            if inspect.iscoroutinefunction(func):
                return AsyncRetry(
                    strategies=strategies,
                    on_exceptions=on_exceptions,
                    before_hooks=before_hooks,
                    after_hooks=after_hooks,
                    retry_exception_hook=retry_exception_hook,
                    should_log=should_log,
                )(func, *args, **kwargs)

            return Retry(
                strategies=strategies,
                on_exceptions=on_exceptions,
                before_hooks=before_hooks,
                after_hooks=after_hooks,
                retry_exception_hook=retry_exception_hook,
                should_log=should_log,
            )(func, *args, **kwargs)

        return t.cast(FuncT[ReturnT], wrapper_retry)

    return decorator_retry



#!/usr/bin/env python3

# Author: Ben Mezger <me@benmezger.nl>
# Created at <2023-09-23 Sat 21:35>

from .exceptions import RetryExaustedError
from .retry import AsyncRetry, BaseRetry, Retry, RetryState, retry
from .strategy import (
    ExponentialBackoffStrategy,
    NoopStrategy,
    SleepStrategy,
    StopAfterAttemptStrategy,
    StopWhenReturnValueStrategy,
    Strategy,
)

__all__ = [
    "AsyncRetry",
    "Retry",
    "BaseRetry",
    "retry",
    "SleepStrategy",
    "StopAfterAttemptStrategy",
    "StopWhenReturnValueStrategy",
    "ExponentialBackoffStrategy",
    "NoopStrategy",
    "Strategy",
    "RetryState",
    "RetryExaustedError",
]



#!/usr/bin/env python3

# Author: Ben Mezger <me@benmezger.nl>
# Created at <2023-09-25 Mon 21:19>


import abc
import logging
import random
import time
import typing as t

from pyretries.exceptions import RetryStrategyExausted

_logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

ConditionT = t.TypeVar("ConditionT")
ReturnT = t.TypeVar("ReturnT")
FuncT = t.Callable[..., ReturnT]
StrategyValueT = t.TypeVar("StrategyValueT")


class Strategy(abc.ABC, t.Generic[StrategyValueT]):
    """Base Strategy class"""

    def __init__(self, should_log: bool = True) -> None:
        """
        Args:
            should_log: Specifies if strategy should log actions
        """
        self.should_log = should_log

    @abc.abstractproperty
    def should_stop(self) -> bool:
        """
        Checks if strategy should apply
        """
        raise NotImplementedError

    @abc.abstractmethod
    def eval(self, value: StrategyValueT | Exception | None) -> bool:
        """
        Evaluates strategy.

        Args:
            value: `func` Returned value or exception.

        Returns:
            `True` if it should still be applied in the next iteration

        Raises:
            RetryStrategyExausted: Raised when strategy is exausted
        """
        raise NotImplementedError


class StopAfterAttemptStrategy(Strategy, t.Generic[StrategyValueT]):
    """
    Stop after attempting N times strategy
    """

    def __init__(self, attempts: int, **kwargs) -> None:
        """
        Args:
            attempts: Number of attempts to run
            kwargs (dict): Passed to base strategy
        """
        self.attempts = attempts
        self.current_attempt = 0
        super().__init__(**kwargs)

    @property
    def should_stop(self) -> bool:
        """Check if strategy should apply.

        Return:
            Returns `True` if current attempt if greater or equal to `attempts`
        """
        if self.current_attempt >= self.attempts:
            return True
        return False

    def eval(self, value: StrategyValueT | Exception | None) -> bool:
        """
        Evaluates strategy.

        Args:
            value: `func` Returned value or exception.

        Return:
            Returns `True` if strategy should be applied in the next iteration

        Raises:
            RetryStrategyExausted: Raised when strategy is exausted
        """
        if self.should_stop:
            raise RetryStrategyExausted from value if isinstance(
                value, Exception
            ) else None

        if self.should_log:
            _logger.info(f"{self.__class__.__name__} is at {self.current_attempt=}")

        self.current_attempt += 1
        return True


class SleepStrategy(Strategy, t.Generic[StrategyValueT]):
    """Sleep strategy"""

    def __init__(self, seconds: float, attempts: int = 1, **kwargs):
        """
        Args:
            seconds: Amount of seconds to sleep
            attempts: Number of maximum attempts
            kwargs (dict): Passed to base strategy
        """
        self.seconds = seconds
        self.attempts = attempts
        self.current_attempt = 0
        super().__init__(**kwargs)

    @property
    def should_stop(self) -> bool:
        """Check if strategy should apply.

        Return:
            Returns `True` if current attempt if greater or equal to `attempts`
        """
        if self.current_attempt >= self.attempts:
            return True
        return False

    def eval(self, value: StrategyValueT | Exception | None) -> bool:
        """
        Evaluates strategy.

        Args:
            value: `func` Returned value or exception.

        Return:
            Returns `True` if strategy should be applied in the next iteration

        Raises:
            RetryStrategyExausted: Raised when strategy is exausted
        """
        if self.should_stop:
            raise RetryStrategyExausted from value if isinstance(
                value, Exception
            ) else None

        self.current_attempt += 1

        if self.should_log:
            _logger.info(
                f"{self.__class__.__name__} {self.current_attempt}/{self.attempts} attempts."
                f" Sleeping for {self.seconds}s"
            )
        time.sleep(self.seconds)

        return True


class NoopStrategy(Strategy, t.Generic[StrategyValueT]):
    "Do nothing strategy"

    def __init__(self, attempts: int = 1, **kwargs) -> None:
        """
        Args:
            attempts: Number of maximum attempts
            kwargs (dict): Passed to base strategy
        """
        self.attempts = attempts
        self.current_attempt = 0
        super().__init__(**kwargs)

    @property
    def should_stop(self) -> bool:
        """Check if strategy should apply.

        Return:
            Returns `True` if current attempt if greater or equal to `attempts`
        """
        if self.current_attempt >= self.attempts:
            return True
        return False

    def eval(self, value: StrategyValueT | Exception | None) -> bool:
        """
        Evaluates strategy.

        Args:
            value: `func` Returned value or exception.

        Return:
            Returns `True` if strategy should be applied in the next iteration

        Raises:
            RetryStrategyExausted: Raised when strategy is exausted
        """
        if self.should_stop:
            raise RetryStrategyExausted from value if isinstance(
                value, Exception
            ) else None

        self.current_attempt += 1
        return True


class StopWhenReturnValueStrategy(Strategy, t.Generic[StrategyValueT]):
    """Stop when return value is X strategy"""

    def __init__(
        self, expected: t.Any, max_attempts: int | None = None, **kwargs
    ) -> None:
        """
        Args:
            expected: Expected return value
            max_attempts: Number of maximum attempts. By default it runs forever
            kwargs (dict): Passed to base strategy
        """
        self.expected = expected
        self.max_attempts = max_attempts
        self.current_attempt = 0
        super().__init__(**kwargs)

    @property
    def should_stop(self) -> bool:
        """Check if strategy should apply.

        Return:
            Returns `True` if current attempt if greater or equal to `attempts`
        """
        if self.max_attempts is not None:
            if self.current_attempt >= self.max_attempts:
                return True
        return False

    def eval(self, value: StrategyValueT | Exception | None) -> bool:
        """
        Evaluates strategy.

        Args:
            value: `func` Returned value or exception.

        Return:
            Returns `True` if strategy should be applied in the next iteration

        Raises:
            RetryStrategyExausted: Raised when strategy is exausted
        """
        if self.should_stop:
            raise RetryStrategyExausted from value if isinstance(
                value, Exception
            ) else None

        if self.max_attempts is not None:
            if self.should_log:
                _logger.info(
                    f"{self.__class__.__name__} is at {self.current_attempt}/{self.max_attempts}."
                )
            self.current_attempt += 1

        return not value == self.expected


class ExponentialBackoffStrategy(Strategy, t.Generic[StrategyValueT]):
    """Exponential backoff strategy"""

    def __init__(self, max_attempts: int, base_delay: float, **kwargs) -> None:
        """
        Args:
            max_attempts: Number of maximum attempts
            base_delay: base delay in seconds for exponential backoff
            kwargs (dict): Passed to base strategy
        """
        self.base_delay = base_delay
        self.max_attempts = max_attempts
        self.current_attempt = 0
        self.delay = 0
        super().__init__(**kwargs)

    @property
    def should_stop(self) -> bool:
        """Check if strategy should apply.

        Return:
            Returns `True` if current attempt if greater or equal to `attempts`
        """
        if self.current_attempt >= self.max_attempts:
            return True
        return False

    def eval(self, value: StrategyValueT | Exception | None) -> bool:
        """
        Evaluates strategy.

        Args:
            value: `func` Returned value or exception.

        Return:
            Returns `True` if strategy should be applied in the next iteration

        Raises:
            RetryStrategyExausted: Raised when strategy is exausted
        """
        if self.should_stop:
            raise RetryStrategyExausted from value if isinstance(
                value, Exception
            ) else None

        self.current_attempt += 1
        self.delay = self.base_delay * 2**self.current_attempt + random.uniform(0, 1)

        if self.should_log:
            _logger.info(
                f"{self.__class__.__name__} {self.current_attempt}/{self.max_attempts} attempts."
                f" Sleeping for {self.delay:.2f}s"
            )

        time.sleep(self.delay)
        return True
